/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* "phase" inital conditions */
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "system.h"

#include "beatbox.h"
#include "state.h"
#include "device.h"
#include "qpp.h"
#include "bikt.h"

#include "k_.h"
#include "mpi_io_choice.h"

extern int Verbose;            /* defined in main */

/* 8< mpi_k_func_rnd_definition */
/* static double _rnd(double a,double b) {return (a+(b-a)*rand()/(RAND_MAX+1.0));} */ // this random number generator is no good, use the MT generator.
/* >8 mpi_k_func_rnd_definition */

#undef SEPARATORS
#define SEPARATORS ";"
#define BLANK " \n\t\r"

typedef struct {
  #include "k_code.h"
  double  *u;		/*[vmax]*/
  real x, y, z;
  int np, nv;
  real *a;		/*[np*nv]*/
  double *p;		/*[nv]*/
  double phaseu, phasep;
  char filename[MAXPATH];
  FILE *file;
  char debugname[MAXPATH];
  FILE *debug;
  p_tb loctb;
  int debugWriter;
} STR;
extern int idev;

// /*#include argsused.h*/
RUN_HEAD(k_func)
  DEVICE_CONST(int, np)
  DEVICE_CONST(int, debugWriter)
  DEVICE_CONST(FILE *,debug)
  DEVICE_CONST(p_tb, loctb)
  int ix, iy, iz, iv, ip, ip1, icode, n, touchPoint;
  double p, q;

  if (debug && debugWriter) {
    fprintf(debug, "t=%ld", (long)t);
    fflush(debug);
  }
  k_on();
  assert(DV==1);
  for (S->z=iz=s.z0;iz<=s.z1;S->z=++iz) {
    for (S->y=iy=s.y0;iy<=s.y1;S->y=++iy) {
      for (S->x=ix=s.x0;ix<=s.x1;S->x=++ix) {
	if (!s.nowhere && debug) {
	  fprintf(debug," (%d %d %d)",(int)ix,(int)iy,(int)iz);
	  fflush(debug);
	}
		
	touchPoint = !s.nowhere && isTissue(ix,iy,iz);
	/* No need to maintain local values if running nowhere. */
	if(touchPoint) memcpy(S->u,New+ind(ix,iy,iz,0),vmax*sizeof(real));
		
	for(icode=0;icode<S->ncode;icode++) {
	  void *code, *result;
	  if NOT(code=(S->code)[icode]) continue;
	  result = execute(code); CHK(NULL);
	  if ((!s.nowhere || debugWriter) && debug) {
	    fprintf(debug," %s=%s",var_name(loctb,(S->data)[icode],n), prt(result, res_type(code)));
	    fflush(debug);
	  }
	  memcpy((S->data)[icode],result,sizetable[res_type(code)]);
		
	  if (((S->data)[icode])==&(S->phasep)) {
	    if (np) {
	      S->phasep = S->phasep*(S->np)*M_1_PI*0.5;
	      while(S->phasep<0) S->phasep=S->phasep+(S->np);
	      while(S->phasep>=(S->np)) S->phasep=S->phasep-(S->np);
	      ip=S->phasep; ip1=(ip+1)%np; p=S->phasep-ip; q=1-p;
	      assert(ip<S->np);
	      for (iv=0;iv<S->nv;iv++){
		S->p[iv]=S->a[iv+S->nv*ip]*q+S->a[iv+S->nv*ip1]*p;
	      }
	    } else {
	      for (iv=0;iv<S->nv;iv++){
		S->p[iv]=S->a[iv+S->nv*ip];
	      }
	    }
	  }
	  
	  if (((S->data)[icode])==&(S->phaseu)) {
	    if (np) {
	      S->phaseu = S->phaseu*(S->np)*M_1_PI*0.5;
	      while(S->phaseu<      0) S->phaseu=S->phaseu+(S->np);
	      while(S->phaseu>=(S->np)) S->phaseu=S->phaseu-(S->np);
	      ip=S->phaseu; ip1=(ip+1)%np; p=S->phaseu-ip; q=1-p;
	      assert(ip<S->np);
	      for (iv=0;iv<S->nv;iv++){
		S->u[iv]=S->a[iv+S->nv*ip]*q+S->a[iv+S->nv*ip1]*p;
	      }
	    } else {
	      for (iv=0;iv<S->nv;iv++) S->u[iv]=S->a[iv+S->nv*ip];
	    }
	  }
	}
	
	if(touchPoint) memcpy(New+ind(ix,iy,iz,0),S->u,vmax*sizeof(real));
      } /* for ix */
    } /* for iy */
  } /* for iz */
  k_off();
  if (debug && debugWriter) {
    fprintf(debug,"\n");
    /* if(fflush_or_not==1) */ fflush(debug); /* debug is about thoroughness, not performance! */
  }
RUN_TAIL(k_func)

DESTROY_HEAD(k_func)
  #include "k_free.h"
  FREE(S->u);
  FREE(S->loctb);
  /* This is the first instance of S->file in k_func.c. */
  /* We can check that it was opened (or at least the pointer defined) somewhere else  */
  /* by a simple hack using ftell. */
  if (S->file && ftell(S->file)!=-1) fclose(S->file);  S->file=NULL;
  if (S->debug && ftell(S->debug)!=-1) fclose(S->debug); S->debug=NULL;  
DESTROY_TAIL(k_func)

CREATE_HEAD(k_func) {
  #define DELIM "\\/,; \r\t\n"
  char *p, name[80];
  int iv, ip, nv, np;
  real *array, *pv;
  
  CALLOC(S->u,vmax,sizeof(double));
	
#if MPI
  /* Identifying the root process allows only one process to write debug output. */
  if (dev->s.nowhere) {
    S->debugWriter = (mpi_rank == 0);
  } else {
    S->debugWriter = (mpi_rank == getRankContainingPoint(dev->s.global_x0,dev->s.global_y0,dev->s.global_z0));
  }
#else
  S->debugWriter = 1;
#endif
	
  ACCEPTF(file,"rt","");
  if(S->file) {
    fgets(buf,MAXSTRLEN,S->file);	/* Count words in the first line */
    if NOT(p=strtok(buf,DELIM)) ERROR1("first line of %s seems empty, cannot recognize",S->filename);
    for(nv=1;NULL!=(p=strtok(NULL,DELIM));nv++);
    rewind(S->file);				/* Count lines in the file */
    for(np=0;fgets(buf,MAXSTRLEN,S->file);np++);
    switch(np) {
    case 0: ERROR1("no lines in %s",S->filename);
    case 1: MESSAGE("/* WARNING: one line found in %s, phase values will be ingored",S->filename);
    }
    rewind(S->file);					/* read the array from file */
    MESSAGE("\n/* Table of %d x %d is assumed in %s */\n",nv,np,S->filename);
    CALLOC(array,nv*np,sizeof(real));
    CALLOC(pv,nv,sizeof(real));
    for (ip=0;ip<np;ip++) {
      if (ip>0){
	for(iv=0;iv<nv;iv++){
	  array[iv+ip*nv]=array[iv+(ip-1)*nv];
	}
      }
      fgets(buf,MAXSTRLEN,S->file);
      p = strtok(buf,DELIM);
      iv=0;
      while (p!=NULL) {
	sscanf(p,REALF,array+iv+ip*nv);
	iv++;
	p = strtok(NULL,DELIM);
      } /*  while */
    } /*  for */
    /* fclose(S->file); */ // this is commented out, as the file is closed elsewhere.
    S->a = array;
    S->np = np;
    S->nv = nv;
    S->p  = pv;
  } else {
    S->a=S->p=NULL;
    S->np=S->nv=0;
  }
  
  ACCEPTF(debug,"wt","");
  k_on();				CHK(NULL);
  memcpy(loctb,deftb,sizeof(*deftb));
  tb_insert_real(loctb,"x",&(S->x));	
  if (err_code) {
    int i;
    for (i=1;i<=maxtab;i++) {
      if (tb_name(deftb,i)[0]=='\0') continue;
      printf("i=%d name=%s\n", i, tb_name(deftb,i));
    }
  }
  CHK("x");
  tb_insert_real(loctb,"y",&(S->y));	CHK("y");
  tb_insert_real(loctb,"z",&(S->z));	CHK("z");
	
/* 8< mpi_k_func_rnd_insertion */
	/*  Random function is safe for local use only. */
	/* tb_insert_fun(loctb,"rnd",_rnd,2);	CHK("rnd"); */
/* >8 mpi_k_func_rnd_insertion */
	
  for(iv=0;iv<vmax;iv++) {
    sprintf(name,"u%d",iv);
    tb_insert_real(loctb,name,&(S->u[iv]));  CHK(name);
  }
	
  if (S->file) {
    tb_insert_real(loctb,"phasep",&(S->phasep));	CHK("phasep");
    tb_insert_real(loctb,"phaseu",&(S->phaseu));	CHK("phaseu");
    for(iv=0;iv<nv;iv++) {
      sprintf(name,"p%d",iv);
      tb_insert_real(loctb,name, &(S->p[iv]));	CHK(name);
    }
  }
	
  if (S->debug) {MALLOC(S->loctb,sizeof(*loctb)); memcpy(S->loctb,loctb,sizeof(*loctb));}
  else S->loctb=NULL;
	
  #include "k_comp.h"
  if (!S->ncode)
    ERROR1("no statements in \"%s\"",buf);
  if (S->file) {
    if(!used(S->data,S->ncode,&(S->phaseu)) && !used(S->data,S->ncode,&(S->phasep))){
      MESSAGE0("/*WARNING: file was specified but \'phase[up]\' not used in the expression!!*/");
    }
  }
  k_off();
  
} CREATE_TAIL(k_func,1)

