/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


/* Write value of specified global k_ variables to file */

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "system.h"
#include "beatbox.h"
#include "device.h"
#include "state.h"
#include "qpp.h"
#include "bikt.h"
#include "windraw.h"
#include "k_.h"
#include "mpi_io_choice.h"

#if defined(IBMPC) || defined(__mips)
#define USEFMT 1
#else
#define USEFMT 0
#endif

typedef struct {
  char filename[MAXPATH];
  FILE *file;
  int append;
  char filehead[80];
  char headformat[80];
  char headcode[80];
  pp_fn headcompiled;
  INT i, N;
  int ncode;
  pp_fn *code;
#if USEFMT
  char format[80];
#endif
  char valuesep[80];
  char fieldsep[80];
  char recordsep[80];
} STR;

#undef SEPARATORS
#define SEPARATORS ";"

/***************/
RUN_HEAD(k_print) {
  DEVICE_CONST(FILE *,file)
  DEVICE_ARRAY(char,headformat)
  DEVICE_CONST(pp_fn,headcompiled)
  DEVICE_CONST(INT,N) DEVICE_VAR(INT,i)
  DEVICE_CONST(int,ncode)
  DEVICE_ARRAY(pp_fn,code)
#if USEFMT
  DEVICE_ARRAY(char,format)
#endif
  DEVICE_ARRAY(char,valuesep)
  DEVICE_ARRAY(char,fieldsep)
  DEVICE_ARRAY(char,recordsep)
  int icode, size;
  char *p;
  
  k_on();
  if (headformat) {
    if (headcompiled)
      fprintf(file,headformat,(real)(*(REAL *)execute(headcompiled)));
    else
      fputs(headformat,file);
  }
  if (ncode) {
    for((*i)=0;(*i)<N;(*i)++) {
      if ((*i)>0) fprintf(file, "%s", fieldsep);
#if USEFMT    
      if (*format) {
	for(icode=0,p=buf;icode<ncode;icode++,p+=size) 
	  memcpy(p,execute(code[icode]),size=res_type(code[icode]));
	vfprintf(file,format,(va_list)buf);
      } else {
#endif
  	for(icode=0;icode<ncode;icode++) {
  	  if (icode) fprintf(file, "%s", valuesep);
	  fprintf(file,"%s",prt(execute(code[icode]),res_type(code[icode])));
  	} /* for */
#if USEFMT
      } /* if format else */
#endif
      } /* for *i */
    } /* if ncode */
    k_off();
      fprintf(file, "%s", recordsep);
    if(fflush_or_not==1)
  fflush(file);

} RUN_TAIL(k_print)

/*********************/
DESTROY_HEAD(k_print) {
  int icode;
  for(icode=0;icode<S->ncode;icode++) FREE(S->code[icode]);
  FREE(S->code);
  FREE(S->headcompiled);
  if (S->file) fclose(S->file); S->file=NULL;
} DESTROY_TAIL(k_print)

/*********************/
CREATE_HEAD(k_print) {
  DEVICE_IS_SPACELESS
  DEVICE_MUST_BE_NOWHERE
	
  ACCEPTI(append,1,0,1);
  ACCEPTF(file,S->append?"at":"wt","stdout");
  ACCEPTS(filehead,"");
  if(*S->filehead) {fprintf(S->file,"%s\n",S->filehead); if(fflush_or_not==1) fflush(S->file);}
  
  k_on();				
  CHK(NULL);
  memcpy(loctb,deftb,sizeof(*deftb));	
  tb_insert_int(loctb,  "i",&(S->i));   CHK("i");
  tb_insert_int(loctb,  "N",&(S->N));   CHK("N");

  ACCEPTS(headformat,"");
  if (S->headformat[0]) { /* need curly bracket as ACCEPTS is a funny macro */
    ACCEPTS(headcode,"");
  } else {
    S->headcode[0]='\0';
  }
  if (S->headcode[0]) {
    S->headcompiled=compile(S->headcode,deftb,t_real); CHK(S->headcode);
  } else {
    S->headcompiled=NULL;
  }
  
  ACCEPTL(N,1,1,LNONE);  
  BEGINBLOCK("list=",buf); {
    int icode;
    char *pcode;
    char *s1=strdup(buf);
    int totalsize=0;
    
    for(S->ncode=0;pcode=strtok(S->ncode?NULL:s1,SEPARATORS);S->ncode+=(*pcode!=0));
    if (!S->ncode) MESSAGE("/*WARNING: no expressions in \"%s\"*/",buf);
    FREE(s1);
    
    if (S->ncode) {  
      if NOT(S->code=calloc(S->ncode,sizeof(pp_fn)))
	      ERROR1("not enough memory for code array of %d",S->ncode);
      
      for(icode=0;icode<S->ncode;icode+=(*pcode!=0)) {
	if NOT(pcode=strtok(icode?NULL:buf,SEPARATORS)) ERROR0("internal error");
	if (!*pcode) continue;
	S->code[icode] = compile(pcode,loctb,t_real); CHK(pcode);
	MESSAGE("\x01\n %s%c",pcode,SEPARATORS[0]);
	totalsize+=sizetable[res_type(S->code[icode])];
      } /* for icode */
      if (totalsize>MAXSTRLEN) ERROR0("MAXSTRLEN not sufficient");
    } /* if S->ncode */
  } ENDBLOCK;

#if USEFMT
  ACCEPTS(format,"");
#endif
  ACCEPTS(valuesep,",");
  ACCEPTS(fieldsep," ");
  ACCEPTS(recordsep,"\n");
  
  k_off();

#if MPI
  /*Since we are writing global k_variables here,
   *	we only need to run this device on one process. */
  if(mpi_rank == 0){
    dev->s.runHere = 1;
  }else{
    dev->s.runHere = 0;
  }
#endif
} CREATE_TAIL(k_print,0)

