/**
 *  Copyright (C) (2010-2012) Vadim Biktashev, Alexander Karpov, 
 *                            Irina Biktasheva, Ross McFarlane;
 *  Copyright (C) 2012 Sanjay Kharche
 *
 *  This file is part of Beatbox.
 *
 *   Beatbox is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Beatbox is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <assert.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "system.h"
#include "beatbox.h"
#include "bikt.h"
#include "device.h"
#include "k_.h"
#include "state.h"
#include "rhs.h"
#include "qpp.h"

#include "sequence.h"
#include "mpi_io_choice.h"

/* 8< mpi_nextq */
int nextq(sequence *q) {
#if MPI
  MPI_Barrier(q->comm);
  if (q->f) {MPI_File_close(&(q->f));	    if (q->mute==0) printf("closed %s\n", q->name);}
#else
  if (q->f) {fclose(q->f);	    if (q->mute==0) printf("closed %s\n", q->name);}
#endif
  if (strchr(q->mask,'%')) {
    sprintf(q->name,q->mask,q->n);
    if (0==strcmp(q->name,q->mask)) ERROR3("bad mask: print %d in format \"%s\" gives \"%s\"",q->n,q->mask,q->name);
#if MPI
    if (MPI_File_open(q->comm, q->name, q->mpiMode, MPI_INFO_NULL, &(q->f)) != MPI_SUCCESS){
      int devRank;
      MPI_Comm_rank(q->comm,&devRank);
      if(devRank==0){
	ERROR1("could not open \"%s\". Check that the directory exists and that you have the necessary permissions.",q->name);
      }else{
	return 0;
      }
    }
#else
    if NOT(q->f=fopen(q->name,q->mode)) ERROR2("cannot open file %s in mode %s. Check that the directory exists and that you have the necessary permissions.", q->name, q->mode);
#endif
  } else {
    char bakname[MAXPATH];
    strcpy(q->name,q->mask);
    sprintf(bakname,"%s.BAK",q->name);
    frename(q->name,bakname);
#if MPI
    if (MPI_File_open(q->comm, q->name, q->mpiMode, MPI_INFO_NULL, &(q->f)) != MPI_SUCCESS){
      int devRank;
      MPI_Comm_rank(q->comm,&devRank);
      if (devRank==0) {
	ERROR1("could not open \"%s\". Check that the directory exists and that you have the necessary permissions.",q->name);
      } else {
	return 0;
      }
    }
#else
    if NOT(q->f=fopen(q->name,q->mode)) ERROR2("cannot open file %s in mode %s. Check that the directory exists and that you have the necessary permissions.", q->name, q->mode);
#endif
  }
  if (!(q->mute)) printf("opened %s\n", q->name);
  (q->n)++;
  return 1;
}

#if MPI
int acceptq (const char *name,const char *mode,const char *deflt,int runHere,sequence *q, char *w) {
  deviceCommunicator(runHere, &(q->comm));
  q->mpiMode = getMPIFileMode(mode);

  if(runHere){
#else
int acceptq (const char *name,const char *mode,const char *deflt,sequence *q, char *w) {
#endif
  char *p;

  q->f = NULL; /*  Prevents error on first close. */

  if NOT(accepts(name,&(q->mask[0]),deflt,w)) return 0;

  if (0==strcmp(q->mask,"") || 0==strcmp(q->mask,null)) {
    strcpy(q->name,null);
    q->f=NULL;
    MESSAGE("\x01%s\"%s\"%c",name,NULLFILE,SEPARATORS[0]);
    return 1;
  } 

  if ((q->mask)[0]=='*') {q->mute=0;memmove((q->mask),(q->mask)+1,MAXPATH-1);} else {q->mute=1;}

  /*  If mask doesn't contain format specifier, test literal filename. */
  if NOT(strchr(q->mask,'%')) {
      strcpy(q->name,q->mask);
      strcpy(q->mode,mode);
#if MPI
      if (MPI_File_open(q->comm, q->name, q->mpiMode, MPI_INFO_NULL, &(q->f)) != MPI_SUCCESS){
	int devRank;
	MPI_Comm_rank(q->comm,&devRank);
	if(devRank==0){
	  ERROR2("could not open \"%s\" in mode \"%s\". Check that the directory exists and that you have the necessary permissions.", q->name,mode);
	}else{
	  return 0;
	}
      }
#else
      if NOT(q->f=fopen(q->name,q->mode)) ERROR2("could not open \"%s\" in mode \"%s\". Check that the directory exists and that you have the necessary permissions.",q->name,q->mode);
#endif
      return 1;
    }

  /*  Mask contains format specifier. */
  strcpy(q->mode,mode);
  strtok(q->mask,",");
  p=strtok(NULL,",");
  if (p==NULL) q->n=0;
  else if (1!=sscanf(p,"%d",&(q->n))) ERROR1("cannot recognise \"%s\" as starting serial number",p);
  
  if (q->n<0) for((q->n)=0;;(q->n)++) {
      sprintf(q->name,q->mask,q->n);
#if MPI
      if (MPI_File_open(q->comm, q->name, MPI_MODE_RDONLY, MPI_INFO_NULL, &(q->f)) != MPI_SUCCESS) break;
      MPI_File_close(&(q->f));
#else
      if NOT((q->f)=fopen(q->name,"r")) break;
      fclose(q->f);
#endif
    }
  if NOT(nextq(q)) {
#if MPI
    int devRank;
    MPI_Comm_rank(q->comm,&devRank);
    if(devRank==0){
#endif
      ERROR3("could not increment \"%s\"=\"%s\",%d",q->name,q->mask,q->n);
#if MPI
    } else {
      return 0;
    }
#endif
  }
  MESSAGE("\x01%s\"%s\",%d=\"%s\"%c",name,q->mask,(q->n)-1,q->name,SEPARATORS[0]);

#if MPI
} /*  if(runHere) */
#endif

  return 1;
}


#if MPI
/*  Returns an MPI file access mode roughly  */
/*  equivalent to the given fopen() mode. */
int getMPIFileMode(char *fm){
  int MPImode;
	
  if (		0==strcmp(fm,"r") || 0==strcmp(fm,"rb")){
    MPImode = MPI_MODE_RDONLY;
    /* MESSAGE("Mode = MPI_MODE_RDONLY");fflush(stdout); // DEBUG */
  } else if (	0==strcmp(fm,"w") || 0==strcmp(fm,"wb") || 
		0==strcmp(fm,"a") || 0==strcmp(fm,"ab") ||
		0==strcmp(fm,"a+")|| 0==strcmp(fm,"ab+")||
		0==strcmp(fm,"a+b")){
    MPImode = MPI_MODE_WRONLY | MPI_MODE_CREATE;
    /* MESSAGE("Mode = MPI_MODE_RDONLY | MPI_MODE_CREATE");fflush(stdout); // DEBUG */
  } else if (	0==strcmp(fm,"r+")|| 0==strcmp(fm,"rb+") || 
		0==strcmp(fm,"r+b")){
    MPImode = MPI_MODE_RDWR;
    /* MESSAGE("Mode = MPI_MODE_RDWR");fflush(stdout); // DEBUG */
  } else if (	0==strcmp(fm,"w+")|| 0==strcmp(fm,"wt") ||
		0==strcmp(fm,"w+b")){
    MPImode = MPI_MODE_RDWR | MPI_MODE_CREATE;
    /* MESSAGE("Mode = MPI_MODE_RDWR | MPI_MODE_CREATE");fflush(stdout); // DEBUG */
  } else{
    ERROR("Could not create equivalent MPI file access mode given \"%s\"\n", fm);
  }
  
  return MPImode;
}
#endif
